import sys

#####
#
# dprint is way easier to read than pprint, but it was originally
# only intended to work well on dicts/lists/scalars.  If it works
# on anything else, then bonus.
#

def dprint( d, indent=0, follow='' ) :
    """ because pprint.PrettyPrinter is too hard to read 

    print nested dictionaries in a nicely indented format
    """
    indent_str='    ' * indent
    l = d.keys()
    l.sort()
    s = indent_str + '{\n'
    maxlen=0
    for x in l :
        if len(x) > maxlen :
            maxlen = len(x)
    maxlen = maxlen + 3
    for x in l :
        s = s + indent_str + ( '%-*s :' % (maxlen,"'%s'"%x) )
        if isinstance(d[x], dict) :
            s = s + "\n" + dprint(d[x], indent+1, follow=',')
        elif isinstance(d[x], list) :
            s = s + "\n" + lprint(d[x], indent+1, follow=',')
        else :
            s = s + " " + repr(d[x])+",\n"
    s = s + indent_str + '}' + follow + '\n'
    return s

def lprint( l, indent=0, follow='' ) :
    indent_str='    ' * indent
    s = indent_str + '[\n'
    for x in l :
        s = s + indent_str
        if isinstance(x, dict) :
            s = s + "\n" + dprint(x, indent+1, follow=',')
        elif isinstance(x, list) :
            s = s + "\n" + lprint(x, indent+1, follow=',')
        else :
            s = s + repr(x)+",\n"
    s = s + indent_str + ']' + follow + '\n'
    return s

#####
#
# read a python expression from a file; useful for creating simple
# data structures for test input.
#

def eval_file( filename ) :
    'open a file, return eval of the contents'
    f=open(filename,"r")
    r = eval(f.read())
    f.close()
    return r

#####
#
# dprint into a file
#

def dlwrite( filename, data, comment = None ):
    ''''write a dict/list item to a file

    The assumption is that the data item is a dict or list, which
    in turn contains only other dicts, lists, or scalars.

    "dl" => "dict / list"
    '''
    f=open(filename,"w")
    if comment is not None :
        f.write('#')
        f.write(comment.replace('\n','\n#'))
        f.write('\n')
    f.write(dprint(data))
    f.close()

#####
#
# return a string showing the current location in the stack.
#

import traceback

def get_stack() :
    return ''.join(traceback.format_stack())


#####
#
# show keys of a nested dict/list structure
#

def print_dict_keys(d, depth = 0) :
    for x in sorted(d.keys()) :
        showitem(x, d[x], depth)

def showitem( name, item, depth ) :
        ty = str(type(item))
        ty = ty.replace('<type ','').replace('>','').replace("'",'')
        if 'numpy.ndarray' in ty :
            ty = ty + ' ' + str( item.shape )+' = ' + str(item.size)
        s = '\t'*depth
        if name != '' :
            s = s + name + ' '
        s = s + ty
        if isinstance(item, dict) :
            print s+'(%d)'%len(item)
            print_dict_keys(item, depth+1)
        elif isinstance(item, tuple) :
            print s+'(%d)' % len(item)
            print_list(item, depth+1)
        elif isinstance(item, list) :
            print s+'(%d)' % len(item)
            print_list(item, depth+1)
        else :
            print s

def print_list(l, depth=0) :
    for x in l :
        showitem( '', x, depth )





"""
import pandokia.helpers.dcomp as dict_comp
import pandokia.helpers.pycode as pycode

# minipyt

    with pycode.test('name') as t :
        okfh = dcomp.open_okfile( t.full_name )
        
        dict_comp.dictionary_comp( 
            data_dict = d, 
            file_base = t.full_name + '.a', 
            t.tda, t.tra,  ... )
        
        dict_comp.dictionary_comp( 
            data_dict = d, 
            file_base = t.full_name + '.b', 
            t.tda, t.tra,  ... )

        okfh.close()

# py.test

    def test_foo( pdk_test_name ) :

        okfh = dcomp.open_okfile( pdk_test_name )
    
        dict_comp.dictionary_comp( ... file_base = pdk_test_name + '.a' ... )

        dict_comp.dictionary_comp( ... file_base = pdk_test_name + '.b' ... )

"""
    

from __future__ import division
import os
import sys
import glob
import re
import errno
import numbers

from   pandokia.helpers.filecomp import safe_rm
import pandokia.helpers.display as display

# looking ahead to python 3
string_type = basestring

##########
#
# The interface to the comparison function gets complicated if the dict
# holds a hierarchical data structure.  Rather than a recursive compare
# and a recursive way to specify tolerances, etc, we flatten the index
# into a single namespace.
#
# The separator is '.' like in python, so a['a.b']=1 can conflict with
# a['a']={'b':1}.  Bummer if you have that kind of structure.
#
# This is a non-destructive operation.
#

class DuplicateIndexException(Exception) :
    "Raised if two different elements of a dict flatten to the same name"


#
# flatten from a dict into another dict; if it is recursive, the prefix
# describes which branch of the tree we are in right now
#
def _flatten_into( from_d, to_d, prefix  ):
    for x in from_d :
        if isinstance( from_d[x], dict ) :
            # A dict is just flattened into the same output dict with
            # a longer prefix.
            _flatten_into( from_d[x], to_d, prefix + x + '.' )

        elif isinstance( from_d[x], ( list, tuple, set ) )  :
            # It is way easier to convert a list into a dict and the
            # flatten the dict.

            # Find out how many digits in the last index value.  Use that
            # to make a format string for the index values.
            fmtlen = len( '%d' % (len(from_d[x])-1) )
            fmtstr = '%%0%dd' % fmtlen

            # actually convert the list to the dict and flatten it
            t = { }
            for i,y in enumerate( from_d[x] ) :
                t[fmtstr % i] = y
            _flatten_into( t, to_d, prefix + x + '.' )

        else :
            # anything else is presumed scalar
            idx = prefix + str(x)
            if idx in to_d :
                raise DuplicateIndexException(idx)
            to_d[ idx ] = from_d[ x ]

def flatten( d ) :
    "Mash a hierarchical dict/list structure (like from json) into a flat dict"
    new = { }
    _flatten_into( d, new, '' )
    return new

##########
# handling of output and reference files
#

class NoReferenceFile(Exception):
    "raise when we cannot find the reference file"

# ensure that required directories exist
for x in ( 'output', 'ref', 'okfile' ) :
    try :
        os.mkdir(x)
    except :
        # No need to write a lot of code to distinguish good/bad exceptions here.
        # If there is a problem, we will find out when we try to use the directory.
        pass

def write_output( file_base, data_dict, interesting_fields ) :
    fn = "output/"+file_base
    safe_rm(fn)
    fp = open(fn,"w")
    dict_copy = {}
    for x in interesting_fields :
        try :
            dict_copy[x] = data_dict[x]
        except KeyError :
            # yes, this means the output file cannot possibly be correct.
            pass
    fp.write(display.dprint(dict_copy))
    fp.close()
    return fn

def read_reference( fn ) :
    try :
        fp = open(fn,"r")
    except IOError, e :
        if e.errno == errno.ENOENT :
            raise NoReferenceFile(fn)
        raise
    d = eval( fp.read() )
    fp.close()
    return d
    

##########


spaces = re.compile("[\s]")

##########
#
# write an okfile for pandokia's flagok feature
#

def open_okfile( test_name ) :
    # the okfile is in the subdirectory okfile/
    okfile = os.getcwd() + '/okfile/' + test_name
    safe_rm( okfile )
    f = open( okfile, "w")
    return f

def append_okfile( okfh, output_file, reference_file ) :
    # we put the okfile in a subdirectory, so the names we have are relative to ../
    if reference_file.startswith('/'):
        okfh.write('../%s %s\n' % ( output_file, reference_file ) )
    else:
        okfh.write('../%s ../%s\n' % ( output_file, reference_file ) )
    okfh.flush()

##########
#
# Compare all the interesting fields.  This function is used by
# the tests.
#

def dictionary_comp(
        data_dict,
            # the actual data
        file_base,
            # implies output file name
            # implies reference file name
            # implies ok file name
            # implies refs/file.fields for compare/ignore fields
            # assumes PDK_REFS environment
        tda = None,
        tra = None,
            # if present, fill tra with output, relative difference
        interesting_fields = None,
            # if interesting, compare only those
        uninteresting_fields = [ ],
            # if uninteresting, exclude those from the default list
        fp_tolerance = 1e-7,
            # 
        tolerance_dict = { },
            # you can give a specific tolerance for specific fields
            # default is no special preference
        print_info = False,
            # 
        okfh = None,
            #
        fail_only = False,
            # if true, only report fields that are involved in a failure
        ) :

    # Start off saying the whole comparison has not failed.  This
    # will be the state until the end, unless some element fails.
    fail = False

    # The list of fields that had an error in the comparison.
    errors = [ ]

    # so we don't have to keep saying "if tda : tda[xx] = yy"
    # these dicts will go out of scope at the end of the function
    if tda is None :
        tda = { }
    if tra is None :
        tra = { }
	
    # a reference dict, generated by hand or from a previous result.
    # or nothing, if it is not there -- gets us through the test

    # BUG: support PDK_REFS
    ref_fn = "ref/"+file_base

    # Read the reference dictionary.  no_reference_file will be the
    # exception if there is no reference file; it is re-raised at
    # the end.  We don't just raise it now because we want all
    # the reporting that comes below.
    try :
        ref_dict = read_reference( ref_fn )
        no_reference_file = None
    except NoReferenceFile as e:
        ref_dict = { }
        no_reference_file = e

    # flatten the data dict (the ref dict is already flattened)
    data_dict = flatten( data_dict )

    # identify interesting fields.  if explicitly specified, use 
    # just that.  Otherwise it defaults to all the fields except
    # those identified as uninteresting.
    if interesting_fields is None :
        interesting_fields = set( [ x for x in data_dict ]
             + [ x for x in ref_dict ] ) - set(uninteresting_fields)

    # save the interesting fields from the flattened dict
    out_fn = write_output( file_base, data_dict, interesting_fields )

    # add a line to the okfile
    if okfh :
        append_okfile( okfh, out_fn, ref_fn )
    
    # Which fields were in the expected answers, but missing from the result.
    missing=set()

    # Which fields did not compare closely enough.
    failed= []

    # Report the default tolerance.
    tda['tolerance'] = fp_tolerance

    # Set up to report the max discrep
    if not fail_only :
        tra['max_discrep']=0
        tra['max_discrep_var']=None

    # Loop over the interesting keys, comparing each one.
    for k in interesting_fields :

        # flag of whether this iteration of the loop fails the compare
        this_fail = False

        # value of the discrepancy for this iteration of the loop
        discrep = None

        # see if there are custom tolerances for this field
        try:
            tol = tolerance_dict[k]
            tda['tol_%s'%k] = tol
        except KeyError:
            tol = fp_tolerance

        # pick out the data value
        try :
            data = data_dict[k]
        except AttributeError :
            tra['discrep_%s'%k] = 'missing data'
            this_fail = True
            missing.add(k)
            continue

        # pick out the reference value
        try :
            ref = ref_dict[k]
        except KeyError :
            tra['discrep_%s'%k] = 'missing ref'
            this_fail = True
            missing.add(k)
            continue

        ## comparing numbers
        if isinstance(data, numbers.Number) and isinstance(ref, numbers.Number) :

            # If both are zero, everything is good.
            if data == 0 and ref == 0 :
                discrep = 0

            else:
                # Compare the discrepancy as a proportion of the reference
                # value.
                try:
                    discrep=(data-ref)/ref
                    if (abs(discrep) > tol):
                        failed.append(k)
                        this_fail = True

                # If the reference is zero, use absolute difference instead.
                except ZeroDivisionError, e:
                    discrep = ( data - ref )
                    if (abs(discrep) > tol):
                        failed.append(k)
                        this_fail = True

        ## comparing strings
        elif isinstance(data, string_type) and isinstance(ref, string_type) :
            if ref.strip() != data.strip() :
                failed.append(k)
                discrep = True
                this_fail = True

        ## comparing non-number, non-string, non-dict, non-list
	    ## If you use json for your data, this just means booleans
	    ## or two different types
        else:
            # neither string nor number - all we have for comparing is equality
            if data != ref :
                failed.append(k)
                discrep = True
                this_fail = True

        # If this element of the data failed, the whole test failed.
        if this_fail :
            fail = True

        # Report this element in the tra if it failed, or if we are
	    # reporting all elements.
        if this_fail or not fail_only :
            # Report the computed value and the reference value as tra.
            tra[k]=str(data)
            tra['ref_%s'%k]=str(ref)
            tra['discrep_%s'%k]=discrep

    # end: for k in interesting_fields

    # If any of the fields failed, report what they were.
    if len(failed) > 0:
        failed.sort()
        tra['failed']=failed

        # For failed numerical values, report the max discrepancy and variable
        discrep, vname = 0.0, None
        for k in failed:
            val=abs(tra['discrep_%s'%k])
            if (isinstance(val,float) and val > discrep):
                discrep=val
                vname=k

        if vname is not None:
            tra['max_discrep']=discrep
            tra['max_discrep_var']=vname


    # anything that wasn't there is also a problem
    if missing :
        m=list(missing)
        m.sort()
        tra['missing']=str(m)
        fail = True

    # This is a pretty display for the interactive user.
    if print_info and ( fail or ( not fail_only ) ) :
        from pandokia.text_table import text_table
        tt = text_table()

        failed_set = set(failed)

        tt.define_column('name')
        tt.define_column('expect')
        tt.define_column('actual')
        tt.define_column('F')
        tt.define_column('discrep')

        for row, k in enumerate( sorted(interesting_fields) ):
            tt.set_value( row, 'name', k )
            if k in ref_dict :
                tt.set_value( row, 'expect', ref_dict[k] )
            if k in data_dict :
                tt.set_value( row, 'actual', data_dict[k] )
            s = 'discrep_%s' % k
            if s in tra :
                tt.set_value( row, 'discrep', tra[s] )
            if k in failed_set :
                tt.set_value( row, 'F', '*' )

        print tt.get_rst(headings=1)

        if missing:
            m=list(missing)
            m.sort()
            print "Missing keys:", m

        if fail :
            print "FAILED"

    if no_reference_file:
        raise no_reference_file

    if len(errors) :
        raise Exception(str(errors))

    if fail :
        raise AssertionError('%s failed' % file_base )




#
# pandokia - a test reporting and execution system
# Copyright 2009, Association of Universities for Research in Astronomy (AURA) 
#
# Assorted functions that are useful in more than one place

import time
import datetime


#
# determine the host name, for reporting purposes
#
# For now, I don't care about the domain because it is always the same
# for me and writing .stsci.edu after the host name all the time just
# causes clutter.  Later, we'll make it configurable whether to
# trim off the domain or not.
#

hostname = None

def gethostname( ) :
    global hostname
    if hostname is None :
        import platform
        hostname = platform.node()
        if "." in hostname :
            hostname = hostname.split(".")[0]
    return hostname
    # bug: decided when to use fqdn or not; override hostname in config; etc



def decode_time_float( istr ) :
    if istr is None :
        return None
    try :
        tyme = float(istr)
    except ValueError :
        if '.' in istr :
            l = istr.split('.',1)
        else :
            l = [ istr ]

        try :
            d = time.strptime( l[0], "%Y-%m-%d %H:%M:%S" )
            tyme = time.mktime(d)
        except ValueError :
            try :
                d = time.strptime( l[0], "%Y-%m-%dT%H:%M:%S" )
                tyme = time.mktime(d)
            except ValueError :
                return None

        if len(l) > 1 :
            a = l[1]
            frac = float(int(a)) / int( '1'+'0'*len(a) )
            tyme = tyme + frac

    return tyme

def decode_time_str( istr ) :
    if istr is None :
        return None
    try :
        tyme = float(istr)
    except ValueError :
        return istr
    return time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(tyme)) + '.' + ("%03d" % ( int( tyme - int(tyme) ) * 1000 ))

def time_diff( max, min ) :
    max = decode_time_float( max )
    min = decode_time_float( min )
    if max is None :
        return None
    if min is None :
        return None
    return datetime.timedelta( seconds= int(max - min ) )



#
# pandokia - a test reporting and execution system
# Copyright 2011, Association of Universities for Research in Astronomy (AURA) 
#
# psycopg database driver
#
# This is almost exactly the mysql driver, but with a few things changed
# for postgres.

__all__ = [ 
    'db_module',
    'db_driver',
    'PandokiaDB',
    'threadsafety',
    ]

import psycopg2 as db_module
import re

# from dbapi.  psycopg2 is level 2 (can use same db connection in
# multiple threads).
threadsafety = db_module.threadsafety


import pandokia.db

# debugging 
_tty = None
# _tty = open("/dev/tty","w")

import cStringIO as StringIO
import os

# use this when something is so specific to the database that you
# can't avoid writing per-database code
db_driver = 'psycopg2'

import re

class PandokiaDB(pandokia.db.where_dict_base) :

    IntegrityError = db_module.IntegrityError
    ProgrammingError = db_module.ProgrammingError
    OperationalError = db_module.OperationalError
    DatabaseError    = db_module.DatabaseError

    # name of this driver.  could be a constant.
    pandokia_driver_name = __module__.split('db_')[1]

    db = None

    def __init__( self, access_arg ) :
        self.db_access_arg = access_arg

    def open( self ) :
        if self.db is None :
            self.db = db_module.connect( ** ( self.db_access_arg ) )
            return

        # bug: see if the connection timed out, and reconnect it

    def start_transaction( self ) :
        if self.db is None :
            self.open()
        self.execute("START TRANSACTION")
        # bug: should we use one the options at
        # http://www.postgresql.org/docs/9.1/static/sql-start-transaction.html

    def commit(self):
        if self.db is None :
            return
        self.db.commit()

    def rollback(self):
        if self.db is None :
            return
        self.db.rollback()

    #
    # explain the query plan using the database-dependent syntax
    #
    def explain_query( self, text, query_dict=None ) :
        if self.db is None :
            self.open()
        f = StringIO.StringIO()
        c = self.execute( 'EXPLAIN '+ text, query_dict )
        for x in c :
            f.write(str(x))
        return f.getvalue()

    #
    # execute a query in a portable way
    # (this capability not offered by dbapi)
    #

    _pat_from = re.compile(':([a-zA-Z0-9_]*)')

    _pat_to = '%(\\1)s '

    def execute( self, statement, parameters = [ ] ) :
        if self.db is None :
            self.open()

        # convert the parameters, as necessary
        if isinstance(parameters, dict) :
            # dict does not need to be converted
            pass
        elif isinstance(parameters, list) or isinstance(parameters, tuple) :
            # list/tuple turned into a dict with string indexes
            tmp = { }
            for x in range(0,len(parameters)) :
                tmp[str(x+1)] = parameters[x]
            parameters = tmp
        elif parameters is None :
            parameters = [ ]
        else :
            # no other parameter type is valid
            raise self.ProgrammingError

        # for mysql, convert :xxx to %(xxx)s
        statement = self._pat_from.sub(self._pat_to, statement)

        # create a cursor, execute the statement
        c = self.db.cursor()

        # print parameters,"<br>"
        c.execute( statement, parameters )

        # return the cursor
        return c

    ## how much table space is this database using
    ## not portable to other DB
    def table_usage( self ) :
        c = self.execute("SELECT pg_database_size( :1 )", ( self.db_access_arg['database'], ) )
        return c.fetchone()[0]

    # 
    def next( self, sequence_name ) :
        if self.db is None :
            self.open()
        c = self.db.cursor()
        c.execute("select nextval('%s')"%sequence_name)
        return c.fetchone()[0]


'''
Ubuntu:

    sudo apt-get install postgresql-client
        gets a client

    sudo apt-get install python-psycopg2
        gets the python client

    sudo apt-get install postgresql
        gets a server

    su postgres

    psql postgres
        \password postgres
        ...enter a password...
            sets password for postgres database user

        create database pandokia;
        create user mark;
        grant all on database pandokia to mark ;

    /etc/postgresql
        config

    /var/lib/postgresql/8.4/main
        data?


Fedora:

    yum install postgresql
        client

    yum install python-psycopg2  
        python client

    yum install postgresql-server
        server

    su postgres
    initdb -D ~postgres/data

    su
    service postgresql start 
    
    su postgres

    psql postgres
        \password postgres
        ...enter a password...
            sets password for postgres database user


Create the database:

    su postgres
    psql postgres
        create database pandokia;
        create user mark;
        grant all on database pandokia to mark ;

        alter user NAME password 'PSWD';

Finding the config file:

    psql postgres
        show config_file;

    "pg_ctl reload" or "kill -HUP server" after editing config
    server is the process running "postmaster"
    (didn't work for me)


psql pandokia
    start the interactive client on the pandokia database


psql
    \d
        like show tables
    \l
        like show databases
    \d tablename
        like describe tablename
    \connect name
        like "use name"


'''



#
# pandokia - a test reporting and execution system
# Copyright 2014, Association of Universities for Research in Astronomy (AURA) 
#
# pymssql database driver - for Microsoft SQL Server
#

###
### WARNING: pymssql can only have _one_ active cursor at a time.  This
### limits the portability of this database layer unless we work out
### a solution.
###
#
# Ideas:
#
# db.execute() returns a cursor.  Each cursor could track if it is
# still in use.  (i.e. It is in use until the __del__ is called.)
# Then when we call db.execute() and a cursor is already active, we
# make a new connection and return a cursor off the new connection.  
# This hoses up transactions, though...
#
# db.execute() returns a cursor.  Internally, we duplicate the
# functions of a cursor and fill it with c.fetchall() before
# returning it.  It may use a lot of memory for the list, but
# if you are aware of this and do not intend to process all
# the data, you might make a shorter query.
#
# db.execute() only has one cursor.  If it determines that the
# cursor is still in use when you run db.execute() again, it
# throws an exception.  Applications have to "del c" to get
# rid of the cursor before db.execute() again.
#


__all__ = [ 
    'db_module',
    'db_driver',
    'PandokiaDB',
    'thread_safe',
    ]

# http://pymssql.org/
# https://pypi.python.org/pypi/pymssql/2.1.0

import pymssql as db_module

# use this when something is so specific to the database that you
# can't avoid writing per-database code
db_driver = 'pymssql'

# from dbapi
thread_safe = db_module.threadsafety

import pandokia.db

# debugging 
_tty = None
# _tty = open("/dev/tty","w")

import os

import re

class PandokiaDB(pandokia.db.where_dict_base) :

    IntegrityError = db_module.IntegrityError
    ProgrammingError = db_module.ProgrammingError
    OperationalError = db_module.OperationalError
    DatabaseError    = db_module.DatabaseError

    # name of this driver.  could be a constant.
    pandokia_driver_name = __module__.split('db_')[1]

    db = None

    def __init__( self, access_arg ) :
        self.db_access_arg = access_arg

    def open( self ) :
        if self.db is None :
            if access_arg.get('password',1) is None :
                raise Exception("password specified as None")
            self.db = db_module.connect( ** ( self.db_access_arg ) )
            return

        # bug: check that the connection is still alive here.

        # wow: http://stackoverflow.com/questions/6179737/sql-server-user-session-time-out
        # "Similarly, a SQL Server session never times out, even if
        # not used for days. The application has to explicitly close the
        # connection for its sessions to terminate."
        #
        # Hmm, but what about a crash / failover in the HA system?

    def start_transaction( self ) :
        if self.db is None :
            self.open()
        self.execute("BEGIN TRANSACTION")

    def commit(self):
        if self.db is None :
            return
        self.db.commit()

    def rollback(self):
        if self.db is None :
            return
        self.db.rollback()

    #
    # execute a query in a portable way
    # (this capability not offered by dbapi)
    #

    _pat_from = re.compile(':([a-zA-Z0-9_]*)')

    _pat_to = '%(\\1)s '

    def execute( self, statement, parameters = [ ], db = None ) :
        if self.db is None :
            self.open()

        # convert the parameters, as necessary
        if isinstance(parameters, dict) :
            # dict does not need to be converted
            pass
        elif isinstance(parameters, list) or isinstance(parameters, tuple) :
            # list/tuple turned into a dict with string indexes
            tmp = { }
            for x in range(0,len(parameters)) :
                tmp[str(x+1)] = parameters[x]
            parameters = tmp
        elif parameters is None :
            parameters = [ ]
        else :
            # no other parameter type is valid
            raise self.ProgrammingError

        # for mysql, convert :xxx to %(xxx)s
        statement = self._pat_from.sub(self._pat_to, statement)

        # create a cursor, execute the statement
        c = self.db.cursor()

        if ( _tty is not None ) and not ( statement.startswith('EXPLAIN') ) :
            _tty.write("--------\nQUERY: %s\nparam %s\n"%(statement,str(parameters)))
            _tty.write(explain_query( statement, parameters ) +"\nWARN: ")
            c.execute("SHOW WARNINGS")
            for x in c :
                _tty.write(str(x)+"\n")
            _tty.write("\n\n\n")

        # print parameters,"<br>"
        c.execute( statement, parameters )

        # return the cursor
        return c

    ## how much table space is this database using, in bytes
    ## not portable to other DB
    def table_usage( self ) :
        c = self.db.cursor()
        c.execute("sp_spaceused")
        n = c.fetchone()
        num, unit = n[1].split()
        if unit == "KB" :
            unit=1024
        elif unit == "MB" :
            unit= 1024*1024
        elif unit == "GB" :
            unit = 1024*1024*1024
        return float(n[0]) * unit


##########
#
"""
MS SQL Server notes


show databases
    select name from master..sysdatabases

show tables
    select * from information_schema.tables

describe tablename
    exec sp_columns tablename

    sp_help @tablename


change password

    ALTER LOGIN username  WITH PASSWORD = 'newpassword' OLD_PASSWORD = 'oldpassword'

"""



import pandokia.helpers.pycode as pycode
import pandokia.helpers.dict_comp as dict_comp
import pandokia.helpers.display as display

dict_a = {
            'a' : 'a',
            'b' : 'b',
            'c' : {
                'x' : 1,
                'y' : 2,
                },
            'd' : [ 10, 11, 12 ],
        }

dict_b = {
        'a' : 'a',
        'b' : 'b',
        'c' : {
            'x' : 1,
            'y' : 2,
            },
        'd' : [ 10, 11, 12 ],
        'e' : [ 20, { 'el2a' : 'el2av', 'el2b' : 'el2bv' }, 21 ],
        }

dict_c = {
        'a' : 'a',
        'b' : 'b',
        'c' : {
            'x' : 1,
            'y' : 2,
            },
        'd' : [ 10, 11, 12 ],
        'e' : [ 20, { 'el2a' : 
                        [ 1, 2, 3 ],
                        'el2b' : 'el2bv' 
                     }, 
                21, 22, 23
            ],
        'f' : [ [ [ 1, 2 ], 2 ], 2 ],
        'g' : { 'a' : { 'b' : { 'c' : 1 } },
                'b' : { 'x' : 1 }
             },
        }

dict_d = dict_c.copy()
dict_d['h'] = [ 1.1, 1.2, 1.2001 ]
 
with pycode.test('flatten') :
    def doit( name, dict_in ) :
        with pycode.test(name) as t :
            dict_out = dict_comp.flatten(dict_in)
            
            print "IN"
            print display.dprint(dict_in)
            print "OUT"
            print display.dprint(dict_out)

            print t.full_name

    doit( 'a', dict_a)

    doit( 'b', dict_b)

    doit( 'c', dict_c)

    with pycode.test('exc') as t :
        try :
            dict_comp.flatten( { 'a' : [ 0, 1 ], 'a.1' : 2, } )
        except dict_comp.DuplicateIndexException :
            pass
        except Exception as e :
            raise
        else :
            assert "did not raise exception"


with pycode.test('compare') :

    with pycode.test('a') as t :
        okfh = dict_comp.open_okfile( t.full_name )
        dict_comp.dictionary_comp( 
            data_dict = dict_a, 
            file_base = t.full_name + '.a', 
            tda = t.tda, tra = t.tra, print_info = True ,
            okfh = okfh)

    with pycode.test('bc') as t :
        okfh = dict_comp.open_okfile( t.full_name )
        dict_comp.dictionary_comp( 
            data_dict = dict_b, 
            file_base = t.full_name + '.b', 
            tda = t.tda, tra = t.tra, print_info = True ,
            okfh = okfh)
        dict_comp.dictionary_comp( 
            data_dict = dict_c, 
            file_base = t.full_name + '.c', 
            tda = t.tda, tra = t.tra, print_info = True ,
            okfh = okfh)

    with pycode.test('d') as t :
        okfh = dict_comp.open_okfile( t.full_name )
        dict_comp.dictionary_comp( 
            data_dict = dict_d, 
            file_base = t.full_name, 
            fp_tolerance = 1e-5,
            tda = t.tda, tra = t.tra, print_info = True,
            okfh = okfh )
